/* interpolation/gsl_interp.h
 * 
 * Copyright (C) 1996, 1997, 1998, 1999, 2000, 2004 Gerard Jungman
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

/* Author:  G. Jungman
 */
#ifndef __GSL_INTERP_H__
#define __GSL_INTERP_H__
#include <stdlib.h>

#define HAVE_INLINE 1

#ifdef HAVE_INLINE
#  if defined(__GNUC_STDC_INLINE__) || defined(GSL_C99_INLINE) || defined(HAVE_C99_INLINE)
#    define INLINE_DECL inline  /* use C99 inline */
#    define INLINE_FUN inline
#  else
#    define INLINE_DECL         /* use GNU extern inline */
#    define INLINE_FUN extern inline
#  endif
#else
#  define INLINE_DECL /* */
#endif

#define RETURN_IF_NULL(x) if (!x) { return ; }

#undef __BEGIN_DECLS
#undef __END_DECLS
#ifdef __cplusplus
# define __BEGIN_DECLS extern "C" {
# define __END_DECLS }
#else
# define __BEGIN_DECLS /* empty */
# define __END_DECLS /* empty */
#endif

__BEGIN_DECLS

#define GSL_VAR extern

enum {
    GSL_SUCCESS  = 0,
    GSL_FAILURE  = -1,
    GSL_CONTINUE = -2,  /* iteration has not converged */
    GSL_EDOM     = 1,   /* input domain error, e.g sqrt(-1) */
    GSL_ERANGE   = 2,   /* output range error, e.g. exp(1e100) */
    GSL_EFAULT   = 3,   /* invalid pointer */
    GSL_EINVAL   = 4,   /* invalid argument supplied by user */
    GSL_EFAILED  = 5,   /* generic failure */
    GSL_EFACTOR  = 6,   /* factorization failed */
    GSL_ESANITY  = 7,   /* sanity check failed - shouldn't happen */
    GSL_ENOMEM   = 8,   /* malloc failed */
    GSL_EBADFUNC = 9,   /* problem with user-supplied function */
    GSL_ERUNAWAY = 10,  /* iterative process is out of control */
    GSL_EMAXITER = 11,  /* exceeded max number of iterations */
    GSL_EZERODIV = 12,  /* tried to divide by zero */
    GSL_EBADTOL  = 13,  /* user specified an invalid tolerance */
    GSL_ETOL     = 14,  /* failed to reach the specified tolerance */
    GSL_EUNDRFLW = 15,  /* underflow */
    GSL_EOVRFLW  = 16,  /* overflow  */
    GSL_ELOSS    = 17,  /* loss of accuracy */
    GSL_EROUND   = 18,  /* failed because of roundoff error */
    GSL_EBADLEN  = 19,  /* matrix, vector lengths are not conformant */
    GSL_ENOTSQR  = 20,  /* matrix not square */
    GSL_ESING    = 21,  /* apparent singularity detected */
    GSL_EDIVERGE = 22,  /* integral or series is divergent */
    GSL_EUNSUP   = 23,  /* requested feature is not supported by the hardware */
    GSL_EUNIMPL  = 24,  /* requested feature not (yet) implemented */
    GSL_ECACHE   = 25,  /* cache limit exceeded */
    GSL_ETABLE   = 26,  /* table limit exceeded */
    GSL_ENOPROG  = 27,  /* iteration is not making progress towards solution */
    GSL_ENOPROGJ = 28,  /* jacobian evaluations are not improving the solution */
    GSL_ETOLF    = 29,  /* cannot reach the specified tolerance in F */
    GSL_ETOLX    = 30,  /* cannot reach the specified tolerance in X */
    GSL_ETOLG    = 31,  /* cannot reach the specified tolerance in gradient */
    GSL_EOF      = 32   /* end of file */
} ;

/* evaluation accelerator */
typedef struct {
  size_t  cache;        /* cache of index   */
  size_t  miss_count;   /* keep statistics  */
  size_t  hit_count;
}
gsl_interp_accel;


/* interpolation object type */
typedef struct {
  const char * name;
  unsigned int min_size;
  void *  (*alloc) (size_t size);
  int     (*init)    (void *, const double xa[], const double ya[], size_t size);
  int     (*eval)    (const void *, const double xa[], const double ya[], size_t size, double x, gsl_interp_accel *, double * y);
  int     (*eval_deriv)  (const void *, const double xa[], const double ya[], size_t size, double x, gsl_interp_accel *, double * y_p);
  int     (*eval_deriv2) (const void *, const double xa[], const double ya[], size_t size, double x, gsl_interp_accel *, double * y_pp);
  int     (*eval_integ)  (const void *, const double xa[], const double ya[], size_t size, gsl_interp_accel *, double a, double b, double * result);
  void    (*free)         (void *);

} gsl_interp_type;


/* general interpolation object */
typedef struct {
  const gsl_interp_type * type;
  double  xmin;
  double  xmax;
  size_t  size;
  void * state;
} gsl_interp;


typedef struct
{
    double * b;
    double * c;
    double * d;
    double * _m;
} akima_state_t;


/* available types */
GSL_VAR const gsl_interp_type * gsl_interp_akima;

gsl_interp_accel *
gsl_interp_accel_alloc(void);

int
gsl_interp_accel_reset (gsl_interp_accel * a);

void
gsl_interp_accel_free(gsl_interp_accel * a);

gsl_interp *
gsl_interp_alloc(const gsl_interp_type * T, size_t n);
     
int
gsl_interp_init(gsl_interp * obj, const double xa[], const double ya[], size_t size);

const char * gsl_interp_name(const gsl_interp * interp);
unsigned int gsl_interp_min_size(const gsl_interp * interp);
unsigned int gsl_interp_type_min_size(const gsl_interp_type * T);


int
gsl_interp_eval_e(const gsl_interp * obj,
                  const double xa[], const double ya[], double x,
                  gsl_interp_accel * a, double * y);

double
gsl_interp_eval(const gsl_interp * obj,
                const double xa[], const double ya[], double x,
                gsl_interp_accel * a);

int
gsl_interp_eval_deriv_e(const gsl_interp * obj,
                        const double xa[], const double ya[], double x,
                        gsl_interp_accel * a,
                        double * d);

double
gsl_interp_eval_deriv(const gsl_interp * obj,
                      const double xa[], const double ya[], double x,
                      gsl_interp_accel * a);

int
gsl_interp_eval_deriv2_e(const gsl_interp * obj,
                         const double xa[], const double ya[], double x,
                         gsl_interp_accel * a,
                         double * d2);

double
gsl_interp_eval_deriv2(const gsl_interp * obj,
                       const double xa[], const double ya[], double x,
                       gsl_interp_accel * a);

int
gsl_interp_eval_integ_e(const gsl_interp * obj,
                        const double xa[], const double ya[],
                        double a, double b,
                        gsl_interp_accel * acc,
                        double * result);

double
gsl_interp_eval_integ(const gsl_interp * obj,
                      const double xa[], const double ya[],
                      double a, double b,
                      gsl_interp_accel * acc);

void
gsl_interp_free(gsl_interp * interp);

size_t
gsl_interp_bsearch(const double x_array[], double x,
                   size_t index_lo, size_t index_hi);

size_t
gsl_interp_accel_find(gsl_interp_accel * a, const double x_array[], size_t size, double x);

__END_DECLS

#endif /* __GSL_INTERP_H__ */
