#include "zls.h"

#define C_KMS 299792.458

int zls_LogXConvolve(Q1DArr wv, Q1DArr fl, Q1DArr cn, Q1DArr nz, Counter nSpec, Real resR, Real fwhm, Real vSini , Real box){


    int err = noErr;

        //
        // spectra are already evenly sampled, and are returned with the same sampling
        // This routine resamples to x5 resolution and log lambda then does
        // convolutions on each of the y arrays, fl5, cn5 and nz5
        // Then it resamples and returns modified arrays.
        // Uses gls akima 3rd order non oscillating splines
        //

    if (( resR <= 0.0 ) && ( fwhm <= 0.0 ) && ( vSini <= 0.0 ) && ( box <= 0.0 )) return noErr; // nothing to do but OK

        // cvt to ln lambda for constant kernels in v space
    Q1DArr lwv = NULL;

    err  = zls_NewQuantity1DArr( &lwv, nSpec );

    if (err == noErr ){ // log lambda array OK

        for (Counter idx = 0; idx < nSpec; idx++) {
            lwv [idx] = log( wv[idx] ); // ln
        }

            // 5 times sampled arrays:

        Q1DArr wv5 = NULL;
        Q1DArr fl5 = NULL;
        Q1DArr cn5 = NULL;
        Q1DArr nz5 = NULL;

        Counter n5 = ((nSpec-1)*5)+1;

        Real ldw  = (lwv[nSpec-1] - lwv[0])/(nSpec-1);
        Real ldw5 = ldw*0.2; // x5 resolution

        err  = zls_NewQuantity1DArr( &wv5, n5 );
        err |= zls_NewQuantity1DArr( &fl5, n5 );
        err |= zls_NewQuantity1DArr( &cn5, n5 );
        err |= zls_NewQuantity1DArr( &nz5, n5 );

        if ( err == noErr ){

            gsl_spline *gSpline = NULL;
            gsl_interp_accel *gSplineAcc = NULL;

            gSplineAcc = gsl_interp_accel_alloc ();
            gSpline = gsl_spline_alloc(gsl_interp_akima, nSpec);

            err = gsl_spline_init (gSpline, lwv, fl, nSpec); // log x axis
            if ( err != noErr ) {
                return err;
            }

            Real w0 = log(wv[ 0 ]); // log lamdda
                                    //Real w0 = (wv[ 0 ]); // log lamdda
            for (Counter idx = 0; idx < n5; idx++) {
                Real x = w0 + idx*ldw5; // uniform in log space
                wv5 [idx] = x;
                Real y = gsl_spline_eval( gSpline, x, gSplineAcc);
                fl5 [idx ] = y;
            }

            err = gsl_spline_init (gSpline, lwv, cn, nSpec);
            if ( err != noErr ) {
                return err;
            }

            for (Counter idx = 0; idx < n5; idx++) {
                Real x = wv5 [idx];
                Real y = gsl_spline_eval( gSpline, x, gSplineAcc);
                cn5 [idx ] = y;
            }

            err = gsl_spline_init (gSpline, lwv, nz, nSpec);
            if ( err != noErr ) {
                return err;
            }

            for (Counter idx = 0; idx < n5; idx++) {
                Real x = wv5 [idx];
                Real y = gsl_spline_eval( gSpline, x, gSplineAcc);
                nz5 [idx ] = y;
            }

            gsl_spline_free(gSpline); gSpline = NULL;
            gsl_interp_accel_free(gSplineAcc); gSplineAcc = NULL;

            /* all specs are x5 and in a uniform log wavelength scale */

            /* convolve with rotation first, then fwhm, then with instrumental resolution */
            /* if resR, fwhm, or vSini <= 0.0 then skip that convolution */

                // do rotation 1st

            if (vSini > 0.0) {

                printf(" Width of Projected Rotation: %#12.5g km/s\n",vSini);

                Real vRes  = C_KMS/vSini;
                Real sigma = log((vRes+vRes+1.0)/(vRes+vRes-1.0));

                err = zls_RotationalConvolution( wv5, fl5, n5, sigma);
                if ( err != noErr ) {
                    return err;
                }

                err = zls_RotationalConvolution( wv5, cn5, n5, sigma);
                if ( err != noErr ) {
                    return err;
                }

                err = zls_RotationalConvolution( wv5, nz5, n5, sigma);
                if ( err != noErr ) {
                    return err;
                }

            }

                // do intrisic gaussian 2nd

            if (fwhm > 0.0) {

                Real gRes = C_KMS/fwhm;
                Real sResR = gRes*2.35482; // FWHM -> sigma
                Real sigma = log((sResR+sResR+1.0)/(sResR+sResR-1.0));

                printf(" FWHM of Gaussian: %#12.5g km/s\n", fwhm);

                err = zls_GaussianConvolution( wv5, fl5, n5, sigma);
                if ( err != noErr ) {
                    return err;
                }

                err = zls_GaussianConvolution( wv5, cn5, n5, sigma);
                if ( err != noErr ) {
                    return err;
                }

                err = zls_GaussianConvolution( wv5, nz5, n5, sigma);
                if ( err != noErr ) {
                    return err;
                }
                
            }
            

            if ( box > 0.0) {

                Real bRes = C_KMS/box;
                Real sResR = bRes*0.5; // FW  -> 0.5
                Real sigma = log((sResR+sResR+1.0)/(sResR+sResR-1.0));

                printf(" FW of Box: %#12.5g km/s\n", box);

                err = zls_BoxcarConvolution( wv5, fl5, n5, sigma);
                if ( err != noErr ) {
                    return err;
                }

                err = zls_BoxcarConvolution( wv5, cn5, n5, sigma);
                if ( err != noErr ) {
                    return err;
                }

                err = zls_BoxcarConvolution( wv5, nz5, n5, sigma);
                if ( err != noErr ) {
                    return err;
                }
                
            }
            
                // do instrumental resolving power last

            if ( resR > 0.0 ){

                    // compute sigma in log ratios
                Real sResR = resR*2.35482; // FWHM -> sigma
                Real sigma = log((sResR+sResR+1.0)/(sResR+sResR-1.0));

                printf(" Resolving Power: %#12.5g\n",resR);
                printf(" FWHM of Instrumental Gaussian: %#12.5g km/s\n",C_KMS/resR);
                printf("Sigma of Instrumental Gaussian is: %#12.5g km/s\n",C_KMS/sResR);

                err = zls_GaussianConvolution( wv5, fl5, n5, sigma);
                if ( err != noErr ) {
                    return err;
                }

                err = zls_GaussianConvolution( wv5, cn5, n5, sigma);
                if ( err != noErr ) {
                    return err;
                }

                err = zls_GaussianConvolution( wv5, nz5, n5, sigma);
                if ( err != noErr ) {
                    return err;
                }

            }

            /* Convert back to original sampling */

            gsl_spline *gReSpline = NULL;
            gsl_interp_accel *gReSplineAcc = NULL;

            gReSplineAcc = gsl_interp_accel_alloc ();
            gReSpline = gsl_spline_alloc(gsl_interp_akima, n5);

            err = gsl_spline_init (gReSpline, wv5, fl5, n5);
            if ( err != noErr ) {
                return err;
            }

            for (Counter idx = 0; idx < nSpec; idx++ ) {
                Real x = lwv [idx];
                Real y = gsl_spline_eval( gReSpline, x, gReSplineAcc);
                fl [idx ] = y;
            }

            err = gsl_spline_init (gReSpline, wv5, cn5, n5);
            if ( err != noErr ) {
                return err;
            }

            for (Counter idx = 0; idx < nSpec; idx++ ) {
                Real x = lwv [idx];
                Real y = gsl_spline_eval( gReSpline, x, gReSplineAcc);
                cn [idx ] = y;
            }

            err = gsl_spline_init (gReSpline, wv5, nz5, n5);
            if ( err != noErr ) {
                return err;
            }

            for (Counter idx = 0; idx < nSpec; idx++ ) {
                Real x = lwv [idx];
                Real y = gsl_spline_eval( gReSpline, x, gReSplineAcc);
                nz [idx ] = y;
            }

            gsl_spline_free(gReSpline); gReSpline = NULL;
            gsl_interp_accel_free(gReSplineAcc); gReSplineAcc = NULL;

            zls_DisposeQuantity1DArr(wv5);
            zls_DisposeQuantity1DArr(fl5);
            zls_DisposeQuantity1DArr(cn5);
            zls_DisposeQuantity1DArr(nz5);

        } // allocated arrays

            // All done, dispose of lwv, and just leave wv untouched.
        zls_DisposeQuantity1DArr(lwv);

    } // alloacted log lambda array

    return (err);
}

int zls_RotationalConvolution( Q1DArr wv, Q1DArr fl, Counter n, Real sigma) {

    int err = noErr;

        // prepare kernel and then do a direct convolution

    Real w0 = wv[0];
    Real w1 = wv[n-1];

    Real dx = (w1-w0)/(n-1); // actually log lambda

    Real invSigma = 1.0/sigma;

    Real cutoff   = sigma/dx; // number of cells in half the resp
    Counter nResp = 2 * (int)ceil(cutoff) + 1; // always odd

    Integer nResp2 = nResp/2;
    Q1DArr resp = NULL;
    err  = zls_NewQuantity1DArr( &resp, nResp );

    if ( err != noErr ) return err;

    Real epsilon = 0.5;

    Real a1 = 0.0;
    Real a2 = 1.0;

    if (epsilon < 1.0) {
        a1=6.283185307179586; // 2pi
        a2=0.5*epsilon/(1.0-epsilon);
    }

    Real respSum = 0.0; // for normalising

        // 3rd order quadrature averaging:

    for (Integer idx = -nResp2; idx <= nResp2; idx++ ){

        Real gm1 = 0.0;
        Real a = (dx*idx - 0.5*dx)*invSigma;
        Real b = 1.0-(a*a);

        if ( b <= 0.0 ) {
            gm1 = 0.0;
        } else {
            gm1 = a1*sqrt(b)+a2*b;//       ! resp = __/\__
        }

        Real g0 = 0.0;
        a = (dx*idx)*invSigma;
        b = 1.0-(a*a);

        if ( b <= 0.0 ) {
            g0 = 0.0;
        } else {
            g0 = a1*sqrt(b)+a2*b; //       ! resp = __/\__
        }

        Real gp1 = 0.0;
        a = (dx*idx + 0.5*dx)*invSigma;
        b = 1.0-(a*a);

        if ( b <= 0.0 ) {
            gp1 = 0.0;
        } else {
            gp1 = a1*sqrt(b)+a2*b; //       ! resp = __/\__
        }

        Real meanG = (gm1 + 4.0*g0 + gp1)/6.0; // 3rd order para quad

        resp[ idx + nResp2 ] = meanG;
        respSum += meanG;

    }

    for (Integer idx = -nResp2; idx <= nResp2; idx++ ){
        resp[ idx + nResp2 ] /= respSum;
    }

        // do the convolution

    err = zls_DirectConvolution  ( fl, resp, n, nResp);

    zls_DisposeQuantity1DArr(resp);

    return err;

}

int zls_GaussianConvolution  ( Q1DArr wv, Q1DArr fl, Counter n, Real sigma) {

    int err = noErr;

    const Real kNSigma = 6.0; // out to 6 sigma

        // prepare gassian kernel and then do a direct convolution

    Real invSigma = 1.0/sigma;

        // make the gaussian response function

    Real w0 = wv[0];
    Real w1 = wv[n-1];

    Real dx = (w1-w0)/(n-1); // actually log lambda

    Real cutoff = kNSigma*sigma/dx; // number of cells in half the resp
    Counter nResp = 2 * (int)floor(cutoff) + 1; // always odd

    Integer nResp2 = nResp/2;

    Q1DArr resp = NULL;

    err  = zls_NewQuantity1DArr( &resp, nResp );
    if ( err != noErr ) return err;

    Real respSum = 0.0; // for normalising

        // 3rd order quadrature averaging:

    for (Integer idx = -nResp2; idx <= nResp2; idx++ ){

        Real a = (dx*idx - 0.5*dx)*invSigma;
        Real gm1 = exp(-0.5*a*a);

        a = (dx*idx)*invSigma;
        Real g0 = exp(-0.5*a*a);

        a = (dx*idx + 0.5*dx)*invSigma;
        Real gp1 = exp(-0.5*a*a);

        Real meanG = (gm1 + 4.0*g0 + gp1)/6.0; // 3rd order para quad

        resp[ idx + nResp2 ] = meanG;
        respSum += meanG;
    }

    for (Integer idx = -nResp2; idx <= nResp2; idx++ ){

        resp[ idx + nResp2 ] /= respSum;

    }
        // do the convolution

    err = zls_DirectConvolution  ( fl, resp, n, nResp);
    
    zls_DisposeQuantity1DArr(resp);
    
    return err;
    
}


int zls_BoxcarConvolution  ( Q1DArr wv, Q1DArr fl, Counter n, Real sigma) {

    int err = noErr;

    const Real kNSigma = 1.0; // out to 1+/-1 HW

        // prepare gassian kernel and then do a direct convolution
        // make the gaussian response function

    Real w0 = wv[0];
    Real w1 = wv[n-1];

    Real dx = (w1-w0)/(n-1); // actually log lambda

    Real cutoff = kNSigma*sigma/dx; // number of cells in half the resp
    Counter nResp = 2 * (int)floor(cutoff) + 1; // always odd

    Integer nResp2 = nResp/2;

    Q1DArr resp = NULL;

    err  = zls_NewQuantity1DArr( &resp, nResp );
    if ( err != noErr ) return err;

    Real respSum = 0.0; // for normalising

        // 3rd order quadrature averaging:

    for (Integer idx = -nResp2; idx <= nResp2; idx++ ){

        Real meanG = 1.0;

        resp[ idx + nResp2 ] = meanG;
        respSum += meanG;
    }

    for (Integer idx = -nResp2; idx <= nResp2; idx++ ){

        resp[ idx + nResp2 ] /= respSum;

    }
        // do the convolution

    err = zls_DirectConvolution  ( fl, resp, n, nResp);
    
    zls_DisposeQuantity1DArr(resp);
    
    return err;
    
}

int zls_DirectConvolution  ( Q1DArr f, Q1DArr resp, Counter n, Counter nResp) {

    int err = noErr;

        // actual convolution done here, on a single prepared evenly sampled
        // input array, response function is -nResp/2 .. 0 .. nResp/2 ie odd
        // 0..nResp, not 0..nResp-1; resp should be normalised to sum to 1.0
        // for simplicty

    Q1DArr out = NULL;

    err  = zls_NewQuantity1DArr( &out, n );

    if ( err != noErr ) return err;

    for (Integer idx = 0; idx < n; idx++ )
        out[ idx ] = 0.0;

        //
        // response function is centered at idx nResp2:
        // |_/\_||
        //

    Integer nResp2 = nResp/2; // ie -10 - 0 - 10 = 21 bins, on array [0..20]
                              // access : (-nResp2 .. nResp2,  + nResp) % nResp  -> 11..20, 0, 1..10

    for ( Integer jdx =-nResp2; jdx <=nResp2; jdx++ ) { // resp loop
        for (Integer idx = 0; idx < nResp2; idx++ ) { // input loop, first half of resp

            Integer l = idx + jdx;
            l = (l < 0)?(-l):(l); // reflection

            Integer kdx = (jdx + nResp2);
            out[idx] += f[l]*resp[kdx];
            
        }
    }
    
    for ( Integer jdx =-nResp2; jdx <=nResp2; jdx++ ) { // resp loop
        for (Integer idx = nResp2; idx < n-nResp2; idx++ ) { // middle input loop
            
            Integer kdx = (jdx + nResp2);
            out[ idx ] += f[idx+jdx]*resp[kdx];
            
        }
    }
    
    for ( Integer jdx = -nResp2; jdx <= nResp2; jdx++ ) { // resp loop
        for (Integer idx = n-nResp2; idx < n; idx++ ) { // input loop, last half of resp
            
            Integer  l=idx+jdx;
            l = (l >= n)?(2*n-l-2):l; // reflection, jump -1 past last cell
            
            Integer kdx = (jdx + nResp2);
            out[ idx ] += f[l]*resp[kdx];
            
        }
    }
    
    for (Integer idx = 0; idx < n; idx++ )
        f[ idx ] = out[ idx ];
    
    zls_DisposeQuantity1DArr(out); out = NULL;
    
    return err;
    
}

